"""Base class for Rule with State Machine."""
import inspect
import os
import pathlib

import HABApp
import HABApp.openhab.connection_handler.func_sync
import transitions.extensions.states

import habapp_rules


@transitions.extensions.states.add_state_features(transitions.extensions.states.Timeout)
class StateMachineWithTimeout(transitions.Machine):
	"""State machine class with timeout"""


@transitions.extensions.states.add_state_features(transitions.extensions.states.Timeout)
class HierarchicalStateMachineWithTimeout(transitions.extensions.HierarchicalMachine):
	"""Hierarchical state machine class with timeout"""


class StateMachineRule(HABApp.Rule):
	"""Base class for creating rules with a state machine."""
	states: list[dict] = []
	trans: list[dict] = []
	state: str

	def __init__(self, state_name: str = None):
		"""Init rule with state machine.

		:param state_name: name of the item to hold the state
		"""
		super().__init__()

		# get prefix for items
		parent_class_path = pathlib.Path(inspect.getfile(self.__class__.__mro__[0]))
		parent_class_path_relative = parent_class_path.relative_to(habapp_rules.BASE_PATH)
		parent_class_path_relative_str = str(parent_class_path_relative).removesuffix(".py").replace(os.path.sep, "_")
		self._item_prefix = f"{parent_class_path_relative_str}.{self.rule_name}".replace(".", "_")

		if not state_name:
			state_name = f"{self._item_prefix}_state"
		self._item_state = self._create_additional_item(state_name, "String")

	@staticmethod
	def _create_additional_item(name: str, item_type: str) -> HABApp.openhab.items.OpenhabItem:
		"""Create additional item if it does not already exists

		:param name: Name of item
		:param item_type: Type of item (e.g. String)
		:return: returns the created item
		"""
		if not HABApp.openhab.interface.item_exists(name):
			HABApp.openhab.interface.create_item(item_type=item_type, name=name, label=name.replace("_", " "))
		return HABApp.openhab.items.OpenhabItem.get_item(name)

	def _get_initial_state(self, default_value: str) -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		if self._item_state.value and self._item_state.value in [item.get("name", None) for item in self.states if isinstance(item, dict)]:
			return self._item_state.value
		return default_value

	def _update_openhab_state(self) -> None:
		"""Update OpenHAB state item. This should method should be set to "after_state_change" of the state machine."""
		self._item_state.oh_send_command(self.state)



"""Rules for bridging KNX controller to MQTT items."""
import logging

import HABApp

habapp_log = logging.getLogger(f"HABApp.{__name__}")


class KnxMqttDimmerBridge(HABApp.Rule):
	"""Create a bridge to control a MQTT dimmer from a KNX controller (e.g. wall switch).

	To use this the items must be configured according the following example:
	- knx_dimmer_ctr: autoupdate must be activated, thing:  [ switch="1/1/124+1/1/120", position="1/1/125+1/1/123", increaseDecrease="1/1/122" ] for switch/position: at first always use the RM-GA, second is the control-GA
	- mqtt_dimmer: autoupdate does not matter, thing: according to OpenHAB documentation

	info: OpenHAB does not support start/stop dimming. Thus, this implementation will set fixed values if INCREASE/DECREASE was received from KNX
	"""

	def __init__(self, knx_dimmer_ctr: str, mqtt_dimmer: str, increase_value: int = 60, decrease_value: int = 30) -> None:
		"""Create object of KNX to MQTT bridge

		:param knx_dimmer_ctr: name of KNX control item
		:param mqtt_dimmer: name of MQTT item
		:param increase_value: value which is set when INCREASE was received.
		:param decrease_value: value which is set when DECREASE was received.
		"""
		HABApp.Rule.__init__(self)

		self.__increase_value = increase_value
		self.__decrease_value = decrease_value

		self._knx_item = HABApp.openhab.items.DimmerItem.get_item(knx_dimmer_ctr)
		self._mqtt_item = HABApp.openhab.items.DimmerItem.get_item(mqtt_dimmer)

		self._knx_item.listen_event(self._cb_knx_event, HABApp.openhab.events.ItemCommandEventFilter())
		self._mqtt_item.listen_event(self._cb_mqtt_event, HABApp.openhab.events.ItemStateChangedEventFilter())

	def _cb_knx_event(self, event: HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Callback, which is called if a KNX command received.

		:param event: HABApp event
		"""
		if isinstance(event.value, (int, float)):
			self._mqtt_item.oh_send_command(event.value)
		elif event.value in {"ON", "OFF"}:
			self._mqtt_item.oh_send_command(event.value)
		elif event.value == "INCREASE":
			if self._mqtt_item.value < self.__increase_value:
				self._mqtt_item.oh_send_command(self.__increase_value)
			else:
				self._mqtt_item.oh_send_command(100)
		elif event.value == "DECREASE":
			if self._mqtt_item.value > self.__decrease_value:
				self._mqtt_item.oh_send_command(self.__decrease_value)
			else:
				self._mqtt_item.oh_send_command(0)
		else:
			habapp_log.error(f"command '{event.value}' ist not supported!")

	def _cb_mqtt_event(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is called if a MQTT state change event happens.

		:param event: HABApp event
		"""
		self._knx_item.oh_post_update(event.value)

		if event.value > 0:
			self._knx_item.oh_post_update("ON")
		else:
			self._knx_item.oh_post_update("OFF")
